Android: add a compat layer for 'syscall.h'
[urcu.git] / urcu.c
diff --git a/urcu.c b/urcu.c
index e6ff0f349e9e4fef17d6d366b5e7197acb841b84..69ebcaa9cb0fbfa07a264ab94fa5261513c72424 100644 (file)
--- a/urcu.c
+++ b/urcu.c
@@ -43,6 +43,7 @@
 #include "urcu/tls-compat.h"
 
 #include "urcu-die.h"
+#include "urcu-wait.h"
 
 /* Do not #define _LGPL_SOURCE to ensure we can emit the wrapper symbols */
 #undef _LGPL_SOURCE
  */
 #define RCU_QS_ACTIVE_ATTEMPTS 100
 
+/*
+ * RCU_MEMBARRIER is only possibly available on Linux.
+ */
+#if defined(RCU_MEMBARRIER) && defined(__linux__)
+#include <urcu/syscall-compat.h>
+#endif
+
+/* If the headers do not support SYS_membarrier, fall back on RCU_MB */
+#ifdef SYS_membarrier
+# define membarrier(...)               syscall(SYS_membarrier, __VA_ARGS__)
+#else
+# define membarrier(...)               -ENOSYS
+#endif
+
+#define MEMBARRIER_EXPEDITED           (1 << 0)
+#define MEMBARRIER_DELAYED             (1 << 1)
+#define MEMBARRIER_QUERY               (1 << 16)
+
 #ifdef RCU_MEMBARRIER
 static int init_done;
 int rcu_has_sys_membarrier;
@@ -82,16 +101,7 @@ void __attribute__((destructor)) rcu_exit(void);
 #endif
 
 static pthread_mutex_t rcu_gp_lock = PTHREAD_MUTEX_INITIALIZER;
-
-int32_t rcu_gp_futex;
-
-/*
- * Global grace period counter.
- * Contains the current RCU_GP_CTR_PHASE.
- * Also has a RCU_GP_COUNT of 1, to accelerate the reader fast path.
- * Written to only by writer with mutex taken. Read by both writer and readers.
- */
-unsigned long rcu_gp_ctr = RCU_GP_COUNT;
+struct rcu_gp rcu_gp = { .ctr = RCU_GP_COUNT };
 
 /*
  * Written to only by each individual reader. Read by both the reader and the
@@ -106,6 +116,12 @@ DEFINE_URCU_TLS(unsigned int, rcu_rand_yield);
 
 static CDS_LIST_HEAD(registry);
 
+/*
+ * Queue keeping threads awaiting to wait for a grace period. Contains
+ * struct gp_waiters_thread objects.
+ */
+static DEFINE_URCU_WAIT_QUEUE(gp_waiters);
+
 static void mutex_lock(pthread_mutex_t *mutex)
 {
        int ret;
@@ -141,7 +157,7 @@ static void mutex_unlock(pthread_mutex_t *mutex)
 static void smp_mb_master(int group)
 {
        if (caa_likely(rcu_has_sys_membarrier))
-               membarrier(MEMBARRIER_EXPEDITED);
+               (void) membarrier(MEMBARRIER_EXPEDITED);
        else
                cmm_smp_mb();
 }
@@ -210,8 +226,8 @@ static void wait_gp(void)
 {
        /* Read reader_gp before read futex */
        smp_mb_master(RCU_MB_GROUP);
-       if (uatomic_read(&rcu_gp_futex) == -1)
-               futex_async(&rcu_gp_futex, FUTEX_WAIT, -1,
+       if (uatomic_read(&rcu_gp.futex) == -1)
+               futex_async(&rcu_gp.futex, FUTEX_WAIT, -1,
                      NULL, NULL, 0);
 }
 
@@ -225,12 +241,12 @@ static void wait_for_readers(struct cds_list_head *input_readers,
        /*
         * Wait for each thread URCU_TLS(rcu_reader).ctr to either
         * indicate quiescence (not nested), or observe the current
-        * rcu_gp_ctr value.
+        * rcu_gp.ctr value.
         */
        for (;;) {
                wait_loops++;
                if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
-                       uatomic_dec(&rcu_gp_futex);
+                       uatomic_dec(&rcu_gp.futex);
                        /* Write futex before read reader_gp */
                        smp_mb_master(RCU_MB_GROUP);
                }
@@ -263,7 +279,7 @@ static void wait_for_readers(struct cds_list_head *input_readers,
                        if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
                                smp_mb_master(RCU_MB_GROUP);
-                               uatomic_set(&rcu_gp_futex, 0);
+                               uatomic_set(&rcu_gp.futex, 0);
                        }
                        break;
                } else {
@@ -282,7 +298,7 @@ static void wait_for_readers(struct cds_list_head *input_readers,
                        if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
                                smp_mb_master(RCU_MB_GROUP);
-                               uatomic_set(&rcu_gp_futex, 0);
+                               uatomic_set(&rcu_gp.futex, 0);
                        }
                        break;
                } else {
@@ -306,9 +322,34 @@ void synchronize_rcu(void)
 {
        CDS_LIST_HEAD(cur_snap_readers);
        CDS_LIST_HEAD(qsreaders);
+       DEFINE_URCU_WAIT_NODE(wait, URCU_WAIT_WAITING);
+       struct urcu_waiters waiters;
+
+       /*
+        * Add ourself to gp_waiters queue of threads awaiting to wait
+        * for a grace period. Proceed to perform the grace period only
+        * if we are the first thread added into the queue.
+        * The implicit memory barrier before urcu_wait_add()
+        * orders prior memory accesses of threads put into the wait
+        * queue before their insertion into the wait queue.
+        */
+       if (urcu_wait_add(&gp_waiters, &wait) != 0) {
+               /* Not first in queue: will be awakened by another thread. */
+               urcu_adaptative_busy_wait(&wait);
+               /* Order following memory accesses after grace period. */
+               cmm_smp_mb();
+               return;
+       }
+       /* We won't need to wake ourself up */
+       urcu_wait_set_state(&wait, URCU_WAIT_RUNNING);
 
        mutex_lock(&rcu_gp_lock);
 
+       /*
+        * Move all waiters into our local queue.
+        */
+       urcu_move_waiters(&waiters, &gp_waiters);
+
        if (cds_list_empty(&registry))
                goto out;
 
@@ -325,10 +366,10 @@ void synchronize_rcu(void)
 
        /*
         * Must finish waiting for quiescent state for original parity before
* committing next rcu_gp_ctr update to memory. Failure to do so could
         * result in the writer waiting forever while new readers are always
         * accessing data (no progress).  Enforce compiler-order of load
* URCU_TLS(rcu_reader).ctr before store to rcu_gp_ctr.
         */
        cmm_barrier();
 
@@ -340,13 +381,13 @@ void synchronize_rcu(void)
        cmm_smp_mb();
 
        /* Switch parity: 0 -> 1, 1 -> 0 */
-       CMM_STORE_SHARED(rcu_gp_ctr, rcu_gp_ctr ^ RCU_GP_CTR_PHASE);
+       CMM_STORE_SHARED(rcu_gp.ctr, rcu_gp.ctr ^ RCU_GP_CTR_PHASE);
 
        /*
* Must commit rcu_gp_ctr update to memory before waiting for quiescent
         * state. Failure to do so could result in the writer waiting forever
         * while new readers are always accessing data (no progress). Enforce
* compiler-order of store to rcu_gp_ctr before load rcu_reader ctr.
         */
        cmm_barrier();
 
@@ -374,6 +415,13 @@ void synchronize_rcu(void)
        smp_mb_master(RCU_MB_GROUP);
 out:
        mutex_unlock(&rcu_gp_lock);
+
+       /*
+        * Wakeup waiters only after we have completed the grace period
+        * and have ensured the memory barriers at the end of the grace
+        * period have been issued.
+        */
+       urcu_wake_all_waiters(&waiters);
 }
 
 /*
@@ -390,6 +438,11 @@ void rcu_read_unlock(void)
        _rcu_read_unlock();
 }
 
+int rcu_read_ongoing(void)
+{
+       return _rcu_read_ongoing();
+}
+
 void rcu_register_thread(void)
 {
        URCU_TLS(rcu_reader).tid = pthread_self();
@@ -460,14 +513,14 @@ void rcu_init(void)
 
 void rcu_exit(void)
 {
-       struct sigaction act;
-       int ret;
-
-       ret = sigaction(SIGRCU, NULL, &act);
-       if (ret)
-               urcu_die(errno);
-       assert(act.sa_sigaction == sigrcu_handler);
-       assert(cds_list_empty(&registry));
+       /*
+        * Don't unregister the SIGRCU signal handler anymore, because
+        * call_rcu threads could still be using it shortly before the
+        * application exits.
+        * Assertion disabled because call_rcu threads are now rcu
+        * readers, and left running at exit.
+        * assert(cds_list_empty(&registry));
+        */
 }
 
 #endif /* #ifdef RCU_SIGNAL */
@@ -325,10 +366,10 @@ void synchronize_rcu(void)
        /*
         * Must finish waiting for quiescent state for original parity before
-        * committing next rcu_gp_ctr update to memory. Failure to do so could
+        * committing next rcu_gp.ctr update to memory. Failure to do so could
         * result in the writer waiting forever while new readers are always
         * accessing data (no progress).  Enforce compiler-order of load
@@ -325,10 +366,10 @@ void synchronize_rcu(void)
         * result in the writer waiting forever while new readers are always
         * accessing data (no progress).  Enforce compiler-order of load
-        * URCU_TLS(rcu_reader).ctr before store to rcu_gp_ctr.
+        * URCU_TLS(rcu_reader).ctr before store to rcu_gp.ctr.
         */
        cmm_barrier();
@@ -340,13 +381,13 @@ void synchronize_rcu(void)
 
        /*
-        * Must commit rcu_gp_ctr update to memory before waiting for quiescent
+        * Must commit rcu_gp.ctr update to memory before waiting for quiescent
         * state. Failure to do so could result in the writer waiting forever
         * while new readers are always accessing data (no progress). Enforce
@@ -340,13 +381,13 @@ void synchronize_rcu(void)
         * state. Failure to do so could result in the writer waiting forever
         * while new readers are always accessing data (no progress). Enforce
-        * compiler-order of store to rcu_gp_ctr before load rcu_reader ctr.
+        * compiler-order of store to rcu_gp.ctr before load rcu_reader ctr.
         */
        cmm_barrier();
This page took 0.025835 seconds and 4 git commands to generate.